#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class BaseTreeNode:
    """
    Node element of a BaseTree
    """

    __slots__ = [
        "node_value",
        "node_value_time",
        "node_type",
        "children",
    ]

    def __init__(self, node_value, node_type, **_):
        self.node_value = node_value
        self.node_value_time = None
        self.node_type = node_type
        self.children = {}

    def set_child(self, key, child):
        """
        Set a child at the given key
        """
        self.children[key] = child

    def pop_child(self, key, default):
        """
        Pop the child the given key
        """
        return self.children.pop(key, default)


class NodeExistsError(Exception):
    """
    Node doesn't exist error
    """


class BaseTree:
    """
    Tree based on BaseTreeNode
    """

    TREE_NODE_CLASS = BaseTreeNode
    __slots__ = ["root"]

    def __init__(self):
        """
        Init the root node
        """
        self.root = self.TREE_NODE_CLASS(None, None)

    def set_node(self, value, node_type, node, timestamp=0):
        """
        Set the node attributes
        Can raise an exception if the node doesn't exists
        :param value: the node 'node_value' attribute to set
        :param node_type: the node 'node_type' attribute to set
        :param node: the node to update
        :param timestamp: the value modification timestamp.
        """
        self._set_node(node, value, node_type, timestamp=timestamp)

    def set_node_at_path(self, value, node_type, path, timestamp=0):
        """
        Set the node attributes
        Creates the node if it doesn't exists
        :param value: the node 'node_value' attribute to set
        :param node_type: the node 'node_type' attribute to set
        :param path: the node path (as a list of string)
        :param timestamp: the value modification timestamp.
        For example:
        - If you created a first node with the path ["my-parent-node"]
        - You can create a child node of my-parent-node by using ["my-parent-node", "my-new-child-node"] as `path`
        :return: void
        """
        self._set_node(
            self.get_or_create_node(path), value, node_type, timestamp=timestamp
        )

    def get_node(self, path, starting_node=None):
        """
        Get the node at the specified path
        :param path: the node path (as a list of string).
        For example:
        - If you created a first node with the path ["my-parent-node"]
        - You can create a child node of my-parent-node by using ["my-parent-node", "my-new-child-node"] as `path`
        :param starting_node: the node to start the relative path
        :return: the node instance or raise a NodeExistsError if the node doesn't exists
        """
        try:
            return self._get_node(path, starting_node=starting_node)
        except KeyError:
            raise NodeExistsError

    def clear(self):
        """
        Clears the whole tree
        """
        self.root = self.TREE_NODE_CLASS(None, None)

    def delete_node(self, path, starting_node=None):
        """
        Delete the node at the specified path
        :param path: the node path (as a list of string).
        For example:
        - If you created a first node with the path ["my-parent-node"]
        - You can create a child node of my-parent-node by using ["my-parent-node", "my-new-child-node"] as `path`
        :param starting_node: the node to start the relative path
        :return: the deleted node or raise a NodeExistsError if the node doesn't exists
        """
        try:
            deleted_node = self._delete_node(path, starting_node=starting_node)
            if deleted_node is None:
                raise NodeExistsError
            return deleted_node
        except KeyError:
            raise NodeExistsError

    def get_or_create_node(self, path, starting_node=None, **kwargs):
        """
        Get the node at the specified path
        Creates the node if it doesn't exists
        :param path: the node path (as a list of string).
        For example:
        - If you created a first node with the path ["my-parent-node"]
        - You can create a child node of my-parent-node by using ["my-parent-node", "my-new-child-node"] as `path`
        :param starting_node: the node to start the relative path
        :return: the node instance
        """
        try:
            return self._get_node(path, starting_node=starting_node)
        except KeyError:
            return self._create_node_path(path, starting_node=starting_node, **kwargs)

    def get_nested_children_with_path(self, path=None, select_leaves_only=True):
        """
        Returns a generator iterating over the nodes children, including nested children. Children are yielded
        together with their node path using a depth-first search (the most nested children are returned first)
        :param path: the path (as a list of string) to the node
        :param select_leaves_only: when True (default), only nodes that don't have children are returned
        :return: a generator of (node, path) tuples
        """
        path = path or []
        return self._get_nested_children_with_path(path, select_leaves_only)

    def _get_nested_children_with_path(self, parent_path, select_leaves_only):
        children_keys = self.get_children_keys(parent_path)
        node = self.get_node(parent_path)
        if not children_keys or not select_leaves_only:
            yield node, parent_path
        for key in children_keys:
            path = list(parent_path)
            path.append(key)
            yield from self._get_nested_children_with_path(path, select_leaves_only)

    def get_children_keys(self, path):
        """
        Return the node's children keys
        Can raise a KeyError if the path does not exists
        :param path: the path (as a list of string) to the node
        :return: children keys as a list
        """
        return list(self.get_node(path).children)

    def _get_node(self, path, starting_node=None):
        """
        Return the node corresponding to the path
        Can raise a KeyError if the path does not exists
        :param path: the path (as a list of string) to the node
        :param starting_node: the node to start the path, root if None
        :return: BaseTreeNode at path
        """
        current_node = self.root if starting_node is None else starting_node
        for key in path:
            current_node = current_node.children[key]
        return current_node

    def _delete_node(self, path, starting_node=None):
        """
        Return the node corresponding to the path
        Can raise a KeyError if the path does not exists
        :param path: the path (as a list of string) to the node
        :param starting_node: the node to start the path, root if None
        :return: BaseTreeNode at path
        """
        current_node = self.root if starting_node is None else starting_node
        for key in path[:-1]:
            current_node = current_node.children[key]
        return current_node.pop_child(path[-1], None)

    def _create_node_path(self, path, starting_node=None, **kwargs):
        """
        Expensive method that creates the path to the selected node
        :param path: path (as a list of string) to the selected node
        :param starting_node: the node to start the path, root if None
        :return: the created node path
        """
        current_node = self.root if starting_node is None else starting_node
        for key in path:
            try:
                current_node = current_node.children[key]
            except KeyError:
                # create a new node as the current node child
                # us it as the new node
                current_node = self.child_factory(current_node, key, **kwargs)

        return current_node

    def child_factory(self, node, key, **kwargs):
        """
        Create a new child an associate it to the given key
        """
        node.set_child(key, self.TREE_NODE_CLASS(None, None, **kwargs))
        return node.children[key]

    def _set_node(self, node, value=None, node_type=None, timestamp=0):
        """
        Sets the node attributes
        :param node: the node instance to update
        :param value: the node instance 'node_value' attribute to set (is ignored if None)
        :param node_type: the node instance 'node_type' attribute to set (is ignored if None)
        :param timestamp: the value modification timestamp.
        """
        if value is not None:
            node.node_value = value

        if node_type is not None:
            node.node_type = node_type

        # set the node value modification timestamp
        node.node_value_time = timestamp



# pylint: disable=R0913,W0718,W0706,C0415
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import uuid

import octobot_commons.profiles.profile_data as profile_data_import
import octobot_commons.logging as bot_logging
import octobot_commons.json_util as json_util
import octobot_commons.constants as constants
import octobot_commons.aiohttp_util as aiohttp_util
import octobot_commons.enums as enums

IMPORTED_AVATAR = "avatar"


async def convert_profile_data_to_profile_directory(
    profile_data: profile_data_import.ProfileData,
    description: str,
    risk: enums.ProfileRisk,
    avatar_url: str,
    output_path: str,
    aiohttp_session,
):
    """
    Creates a profile folder from the given ProfileData
    :param profile_data: path to the profile zipped archive
    :param description: profile description
    :param risk: profile risk
    :param avatar_url: profile avatar_url
    :param output_path: profile folder path
    :param aiohttp_session: session to use
    """
    logger = bot_logging.get_logger(__name__)
    if os.path.exists(output_path):
        raise OSError(f"{output_path} already exists")
    os.mkdir(output_path)
    profile = _get_profile(profile_data, description, risk, output_path)
    # tentacles_config.json
    tentacles_setup_config = _get_tentacles_setup_config(profile_data, output_path)
    tentacles_setup_config.save_config()
    # specific_config
    _save_specific_config(profile_data, output_path)
    # avatar file
    try:
        await _download_and_set_avatar(
            profile, avatar_url, output_path, aiohttp_session
        )
    except Exception as err:
        logger.exception(err, True, f"Error when downloading profile avatar: {err}")
    # finish with profile.json to include edits from previous methods
    profile.save()


def _get_profile(
    profile_data: profile_data_import.ProfileData,
    description: str,
    risk: enums.ProfileRisk,
    output_path: str,
):
    profile = profile_data.to_profile(output_path)
    # use trading simulator by default
    profile.config[constants.CONFIG_TRADER][constants.CONFIG_ENABLED_OPTION] = False
    profile.config[constants.CONFIG_SIMULATOR][constants.CONFIG_ENABLED_OPTION] = True
    profile.description = description
    profile.risk = risk
    profile.profile_id = str(uuid.uuid4().hex)
    profile.read_only = True
    return profile


def _get_tentacles_setup_config(
    profile_data: profile_data_import.ProfileData, output_path: str
):
    try:
        import octobot_tentacles_manager.api

        classes = [
            octobot_tentacles_manager.api.get_tentacle_class_from_string(
                tentacle_data.name
            ).get_name()
            for tentacle_data in profile_data.tentacles
        ]
        config_path = os.path.join(output_path, constants.CONFIG_TENTACLES_FILE)
        tentacles_setup_config = (
            octobot_tentacles_manager.api.create_tentacles_setup_config_with_tentacles(
                *classes, config_path=config_path
            )
        )
        octobot_tentacles_manager.api.fill_with_installed_tentacles(
            tentacles_setup_config
        )
        return tentacles_setup_config
    except ImportError:
        raise


def _save_specific_config(
    profile_data: profile_data_import.ProfileData, output_path: str
):
    try:
        import octobot_tentacles_manager.constants

        specific_config_dir = os.path.join(
            output_path,
            octobot_tentacles_manager.constants.TENTACLES_SPECIFIC_CONFIG_FOLDER,
        )
        os.mkdir(specific_config_dir)
        for tentacle_config in profile_data.tentacles:
            json_util.safe_dump(
                tentacle_config.config,
                os.path.join(
                    specific_config_dir,
                    f"{tentacle_config.name}{octobot_tentacles_manager.constants.CONFIG_EXT}",
                ),
            )
    except ImportError:
        raise


async def _download_and_set_avatar(
    profile, avatar_url, output_path: str, aiohttp_session
):
    profile.avatar = _get_avatar_name(avatar_url)
    try:
        import aiofiles

        async with aiofiles.open(
            os.path.join(output_path, profile.avatar), "wb+"
        ) as downloaded_file:
            await aiohttp_util.download_stream_file(
                downloaded_file,
                avatar_url,
                aiohttp_session,
                is_aiofiles_output_file=True,
            )
    except ImportError:
        raise


def _get_avatar_name(avatar_url: str):
    # remove params and get file name with ext
    return avatar_url.split("?")[0].split("/")[-1]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid
import octobot_commons.configuration as configuration


def test_get_password_hash():
    assert len(configuration.get_password_hash("")) == 64
    assert len(configuration.get_password_hash("1")) == 64
    assert len(configuration.get_password_hash("1a")) == 64
    for _ in range(100):
        rand_password = str(uuid.uuid4())
        hashed_password = configuration.get_password_hash(rand_password)
        assert len(hashed_password) == 64
        assert not hashed_password == rand_password



# pylint: disable=W0703
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time

import octobot_commons.enums as enums
import octobot_commons.logging as logging
import octobot_commons.databases.implementations as databases_implementations
import octobot_commons.databases.run_databases.utils as run_databases_utils


class AbstractRunDatabasesPruner:
    def __init__(self, run_databases_identifier, max_databases_size):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.database_adaptor = run_databases_identifier.database_adaptor
        self.databases_root_identifier = run_databases_identifier.data_path
        self.max_databases_size = max_databases_size
        self.all_db_data = []
        self.backtesting_run_path_identifier = (
            run_databases_utils.get_backtesting_related_run_path_identifiers_str(
                self.database_adaptor
            )
        )
        self._run_db = run_databases_identifier.get_db_full_name(
            enums.RunDatabases.RUN_DATA_DB.value
        )

    async def explore(self):
        """
        Explore self.databases_root_identifier to gather storage
        statistics to be used in prune_oldest_run_databases
        """
        t_start = time.time()
        await self._explore_databases()
        total_time = round(time.time() - t_start, 2)
        if total_time > 1:
            self.logger.debug(
                f"Explored run databases for pruning in {total_time} seconds."
            )

    async def prune_oldest_run_databases(self):
        """
        Delete the necessary backtesting run data for the total backtesting storage
        size to be <= self.max_databases_size. Deletes oldest run data first
        """
        self.all_db_data = sorted(
            self.all_db_data, key=lambda data: data.last_modified_time
        )
        removed_databases = []
        while self._get_total_db_size() > self.max_databases_size:
            if await self._prune_database(self.all_db_data[0]):
                removed_databases.append(self.all_db_data[0])
                self.all_db_data = self.all_db_data[1:]
        if removed_databases:
            await self._update_backtesting_runs_metadata(removed_databases)
            self._log_summary(removed_databases)

    async def _explore_databases(self):
        raise NotImplementedError("_explore_databases is not implemented")

    async def _prune_database(self, db_data):
        raise NotImplementedError("_prune_database is not implemented")

    async def _get_global_runs_identifiers(self, removed_databases):
        raise NotImplementedError("_get_global_runs_identifiers is not implemented")

    async def _update_backtesting_runs_metadata(self, removed_databases):
        for global_runs_identifier in await self._get_global_runs_identifiers(
            removed_databases
        ):
            await self._update_metadata(global_runs_identifier)

    async def _update_metadata(self, global_runs_identifier):
        run_db_identifier = run_databases_utils.get_global_run_database_identifier(
            global_runs_identifier
        )
        if run_db_identifier is not None:
            remaining_run_ids = {
                int(identifier)
                for identifier in await run_db_identifier.get_backtesting_run_ids()
            }
            async with databases_implementations.DBWriterReader.database(
                run_db_identifier.get_backtesting_metadata_identifier()
            ) as reader_writer:
                found_runs = await reader_writer.all(enums.DBTables.METADATA.value)
                # iterate in reverse order to keep only latest appearance of each id
                metadata = []
                added_runs = set()
                for run in found_runs[::-1]:
                    run_id = run[enums.BacktestingMetadata.ID.value]
                    if run_id in remaining_run_ids and run_id not in added_runs:
                        metadata.append(run)
                        added_runs.add(run_id)
                await reader_writer.replace_all(enums.DBTables.METADATA.value, metadata)

    def _log_summary(self, removed_databases):
        first_removed = removed_databases[0].get_human_readable_last_modified_time()
        last_removed = removed_databases[-1].get_human_readable_last_modified_time()
        self.logger.debug(
            f"Deleted the {len(removed_databases)} oldest run data from the {first_removed} to the {last_removed}"
        )

    def _get_total_db_size(self):
        return sum(db_data.size for db_data in self.all_db_data)


class DBData:
    def __init__(self, identifier, parts):
        self.identifier = identifier
        self.parts = parts
        self.size = sum(part.size for part in self.parts)
        self.last_modified_time = max(part.last_modified_time for part in self.parts)

    def get_human_readable_last_modified_time(self):
        """
        :return: self.last_modified_time in a human-readable format
        """
        return time.strftime(
            "%Y-%m-%d %H:%M:%S", time.strptime(time.ctime(self.last_modified_time))
        )


class AbstractDBPartData:
    def __init__(self, identifier):
        self.identifier = identifier
        self.size = None
        self.last_modified_time = None



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.databases as databases


def test_remove_oldest_elements():
    databases.GlobalSharedMemoryStorage.instance()["10"] = "a"
    databases.GlobalSharedMemoryStorage.instance()["2"] = 1
    databases.GlobalSharedMemoryStorage.instance()["-1"] = "af"
    databases.GlobalSharedMemoryStorage.instance()["0"] = "as"
    databases.GlobalSharedMemoryStorage.instance()[10] = "asfgvfs"
    databases.GlobalSharedMemoryStorage.instance()["12"] = "bbcc"
    assert len(databases.GlobalSharedMemoryStorage.instance()) == 6
    assert databases.GlobalSharedMemoryStorage.instance() == {
        "10": "a",
        "2": 1,
        "-1": "af",
        "0": "as",
        10: "asfgvfs",
        "12": "bbcc",
    }
    databases.GlobalSharedMemoryStorage.instance().remove_oldest_elements(0)
    assert databases.GlobalSharedMemoryStorage.instance() == {
        "10": "a",
        "2": 1,
        "-1": "af",
        "0": "as",
        10: "asfgvfs",
        "12": "bbcc",
    }
    databases.GlobalSharedMemoryStorage.instance().remove_oldest_elements(1)
    assert databases.GlobalSharedMemoryStorage.instance() == {
        "2": 1,
        "-1": "af",
        "0": "as",
        10: "asfgvfs",
        "12": "bbcc",
    }
    databases.GlobalSharedMemoryStorage.instance().remove_oldest_elements(4)
    assert databases.GlobalSharedMemoryStorage.instance() == {
        "12": "bbcc",
    }
    databases.GlobalSharedMemoryStorage.instance()["2"] = 1
    assert databases.GlobalSharedMemoryStorage.instance() == {
        "12": "bbcc",
        "2": 1,
    }
    databases.GlobalSharedMemoryStorage.instance().remove_oldest_elements(1)
    assert databases.GlobalSharedMemoryStorage.instance() == {
        "2": 1,
    }
    databases.GlobalSharedMemoryStorage.instance().remove_oldest_elements(10)
    assert databases.GlobalSharedMemoryStorage.instance() == {}


def test_get_bytes_size():
    assert 0 < databases.GlobalSharedMemoryStorage.instance().get_bytes_size() < 1000
    for i in range(10000):
        databases.GlobalSharedMemoryStorage.instance()[i] = "aaaaaaaaaaaaaaaaaaaaaa"
    assert 200000 < databases.GlobalSharedMemoryStorage.instance().get_bytes_size() < 800000
    databases.GlobalSharedMemoryStorage.instance().remove_oldest_elements(10000)



# pylint: disable=R0913,W0703
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import os
import zipfile
import shutil
import pathlib
import uuid
import time
import requests
import jsonschema
import octobot_commons.constants as constants
import octobot_commons.enums as enums
import octobot_commons.logging as bot_logging
import octobot_commons.errors as errors
import octobot_commons.json_util as json_util

# avoid cyclic import
from octobot_commons.profiles.profile import Profile
import octobot_commons.profiles.profile_data as profile_data_import
import octobot_commons.profiles.profile_data_import as profile_data_importer


NON_OVERWRITTEN_PROFILE_FOLDERS = []
NON_OVERWRITTEN_PROFILE_FILES = [constants.PROFILE_CONFIG_FILE]
try:
    import octobot_tentacles_manager.constants as tentacles_manager_constants

    NON_OVERWRITTEN_PROFILE_FOLDERS.append(
        tentacles_manager_constants.TENTACLES_SPECIFIC_CONFIG_FOLDER
    )
except ImportError:
    pass


def export_profile(profile, export_path: str) -> str:
    """
    Exports the given profile into export_path, appends ".zip" as a file extension
    :param profile: profile to export
    :param export_path: export path ending with filename
    :return: the exported profile path including file extension
    """
    temp_path = f"{export_path}{int(time.time() * 1000)}"
    # remove any existing file to prevent any side effect
    if os.path.exists(temp_path):
        raise OSError(f"Can't export profile, the {temp_path} folder exists")
    export_path_with_ext = f"{export_path}.{constants.PROFILE_EXPORT_FORMAT}"
    if os.path.isfile(export_path_with_ext):
        os.remove(export_path_with_ext)
    # copy profile into a temp dir to edit it
    shutil.copytree(profile.path, temp_path)
    try:
        _filter_profile_export(temp_path)
        # export the edited profile
        shutil.make_archive(
            os.path.abspath(export_path), constants.PROFILE_EXPORT_FORMAT, temp_path
        )
    finally:
        shutil.rmtree(temp_path)
    return export_path_with_ext


def install_profile(
    import_path: str,
    profile_name: str,
    bot_install_path: str,
    replace_if_exists: bool,
    is_imported: bool,
    origin_url: str = None,
    quite: bool = False,
    profile_schema: str = None,
) -> Profile:
    """
    Installs the given profile export archive into the user's profile directory
    :param import_path: path to the profile zipped archive
    :param profile_name: name of the profile folder
    :param bot_install_path: path to the octobot installation
    :param replace_if_exists: when True erase the profile with the same name if it exists
    :param is_imported: when True the profile is set as imported
    :param origin_url: url the profile is coming from (if relevant)
    :param quite: when True, only log errors
    :param profile_schema: the schema to validate profile against
    :return: The created profile
    """
    logger = bot_logging.get_logger("ProfileSharing")
    target_import_path = _get_target_import_path(
        bot_install_path, profile_name, replace_if_exists
    )
    action = "Creat"
    if replace_if_exists:
        action = "Updat"
    if not quite:
        logger.info(f"{action}ing {profile_name} profile.")
    _import_profile_files(import_path, target_import_path)
    profile = Profile(target_import_path, schema_path=profile_schema).read_config()
    profile.imported = is_imported
    profile.origin_url = origin_url
    _ensure_unique_profile_id(profile)
    if is_imported:
        try:
            profile.validate()
        except jsonschema.exceptions.ValidationError as err:
            shutil.rmtree(target_import_path)
            raise errors.ProfileImportError(
                f"Invalid imported profile: {err.message} in '{'/'.join(err.absolute_path)}'"
            ) from err
    profile.save()
    if not quite:
        logger.info(f"{action}ed {profile.name} ({profile_name}) profile.")
    return profile


def import_profile(
    import_path: str,
    profile_schema: str,
    name: str = None,
    bot_install_path: str = ".",
    origin_url: str = None,
) -> Profile:
    """
    Imports the given profile export archive into the user's profile directory with the "imported_" prefix
    :param import_path: path to the profile zipped archive
    :param profile_schema: the schema to validate profile against
    :param name: name of the profile folder
    :param bot_install_path: path to the octobot installation
    :param origin_url: url the profile is coming from
    :return: The created profile
    """
    temp_profile_name = _get_profile_name(name, import_path)
    profile = install_profile(
        import_path,
        temp_profile_name,
        bot_install_path,
        False,
        True,
        origin_url=origin_url,
        profile_schema=profile_schema,
    )
    if profile.name != temp_profile_name:
        profile.rename_folder(_get_unique_profile_folder_from_name(profile), False)
    return profile


async def import_profile_data_as_profile(
    profile_data: profile_data_import.ProfileData,
    profile_schema: str,
    aiohttp_session,
    name: str = None,
    description: str = None,
    risk: enums.ProfileRisk = None,
    bot_install_path: str = ".",
    origin_url: str = None,
    logo_url: str = None,
) -> Profile:
    """
    Imports the given ProfileData into the user's profile directory with the "imported_" prefix
    :param profile_data: path to the profile zipped archive
    :param profile_schema: the schema to validate profile against
    :param name: name of the profile folder
    :param bot_install_path: path to the octobot installation
    :param origin_url: url the profile is coming from
    :param logo_url: url the profile avatar
    :return: The created profile
    """
    logger = bot_logging.get_logger("ProfileSharing")
    import_path = f"{name}-{uuid.uuid4().hex}"
    try:
        profile_data.profile_details.name = name
        await profile_data_importer.convert_profile_data_to_profile_directory(
            profile_data, description, risk, logo_url, import_path, aiohttp_session
        )
        return import_profile(
            import_path=import_path,
            profile_schema=profile_schema,
            name=name,
            bot_install_path=bot_install_path,
            origin_url=origin_url,
        )
    finally:
        try:
            if os.path.isdir(import_path):
                shutil.rmtree(import_path)
        except Exception as err:
            logger.exception(err, True, f"Error when removing profile temp dir: {err}")


def download_profile(url, target_file, timeout=60):
    """
    Downloads a profile from the given url
    :param url: profile url
    :param target_file: path to save the file
    :param timeout: time given to the request before timeout
    :return: saved file path
    """
    # unauthenticated download
    with requests.get(url, stream=True, timeout=timeout) as req:
        req.raise_for_status()
        with open(target_file, "wb") as write_file:
            for chunk in req.iter_content(chunk_size=8192):
                write_file.write(chunk)
    return target_file


def download_and_install_profile(download_url, profile_schema):
    """
    :param download_url: profile url
    :param profile_schema: the schema to validate profile against
    :return: the installed profile, None if an error occurred
    """
    logger = bot_logging.get_logger("ProfileSharing")
    name = download_url.split("/")[-1]
    file_path = None
    try:
        file_path = download_profile(download_url, name)
        profile = import_profile(
            file_path, profile_schema, name=name, origin_url=download_url
        )
        logger.info(
            f"Downloaded and installed {profile.name} from {profile.origin_url}"
        )
        return profile
    except errors.UnsupportedError as err:
        logger.error(f"Error when installing profile: {err}")
        return None
    except Exception as err:
        logger.exception(err, True, f"Error when installing profile: {err}")
        return None
    finally:
        if file_path is not None and os.path.isfile(file_path):
            os.remove(file_path)


def _get_profile_name(name, import_path):
    profile_name = name or (
        f"{constants.IMPORTED_PROFILE_PREFIX}_{os.path.split(import_path)[-1]}"
    )
    return profile_name.split(f".{constants.PROFILE_EXPORT_FORMAT}")[0]


def _filter_profile_export(profile_path: str):
    profile_file = os.path.join(profile_path, constants.PROFILE_CONFIG_FILE)
    if os.path.isfile(profile_file):
        parsed_profile = json_util.read_file(profile_file)
        _filter_disabled(parsed_profile, constants.CONFIG_EXCHANGES)
        with open(profile_file, "w") as open_file:
            json.dump(parsed_profile, open_file, indent=4, sort_keys=True)


def _filter_disabled(profile_config: dict, element):
    filtered_exchanges = {
        exchange: details
        for exchange, details in profile_config[constants.PROFILE_CONFIG][
            element
        ].items()
        if details.get(constants.CONFIG_ENABLED_OPTION, True)
    }
    profile_config[constants.PROFILE_CONFIG][element] = filtered_exchanges


def _get_target_import_path(
    bot_install_path: str, profile_name: str, replace_if_exists: bool
) -> str:
    """
    Get the target profile folder path
    :param bot_install_path: path to the octobot installation
    :param profile_name: name of the profile folder
    :param replace_if_exists: when True erase the profile with the same name if it exists
    :return: (the final target import path, True if the profile is replaced)
    """
    target_import_path = os.path.join(
        bot_install_path, constants.USER_PROFILES_FOLDER, profile_name
    )
    if replace_if_exists:
        return target_import_path
    return _get_unique_profile_folder(target_import_path)


def _import_profile_files(profile_path: str, target_profile_path: str) -> None:
    """
    Copy or extract profile files to destination. Does not override local tentacles configuration
    :param profile_path: the current profile path
    :param target_profile_path: the target profile path
    :return: None
    """
    if zipfile.is_zipfile(profile_path):
        with zipfile.ZipFile(profile_path) as zipped_profile:
            for archive_member in zipped_profile.namelist():
                if _should_profile_file_be_imported(
                    target_profile_path, archive_member
                ):
                    zipped_profile.extract(archive_member, target_profile_path)
    else:
        if not os.path.isdir(profile_path):
            raise errors.UnsupportedError(
                f"Profile format not supported ({profile_path})"
            )

        def _get_ignored_elements(current_dir, sub_elements):
            return [
                element
                for element in sub_elements
                if not _should_profile_file_be_imported(
                    target_profile_path,
                    os.path.join(current_dir, element).replace(
                        f"{profile_path}{os.path.sep}", ""
                    ),  # force local path
                )
            ]

        shutil.copytree(
            profile_path,
            target_profile_path,
            ignore=_get_ignored_elements,
            dirs_exist_ok=True,
        )


def _should_profile_file_be_imported(
    target_profile_path: str, profile_file_path: str
) -> bool:
    for non_overwritten_element in NON_OVERWRITTEN_PROFILE_FOLDERS:
        # ignore files in NON_OVERWRITTEN_PROFILE_FOLDERS that already exist
        element_path = pathlib.Path(profile_file_path)
        if (
            element_path.name in NON_OVERWRITTEN_PROFILE_FILES
            or non_overwritten_element in element_path.parts[:-1]
        ) and os.path.isfile(os.path.join(target_profile_path, profile_file_path)):
            return False
    return True


def _get_unique_profile_folder_from_name(profile) -> str:
    folder = _get_unique_profile_folder(
        os.path.join(os.path.split(profile.path)[0], profile.name)
    )
    return os.path.split(folder)[1]


def _get_unique_profile_folder(target_import_path: str) -> str:
    """
    Creates an unique profile folder name
    :param target_import_path: the expected target profile folder name
    :return: the unique profile folder name
    """
    iteration = 1
    candidate = target_import_path
    while os.path.exists(candidate) and iteration < 100:
        iteration += 1
        candidate = f"{target_import_path}_{iteration}"
    return candidate


def _ensure_unique_profile_id(profile) -> None:
    """
    Ensure that no other installed profile has the same id
    :param profile: the installed profile
    :return: None
    """
    ids = Profile.get_all_profiles_ids(
        pathlib.Path(profile.path).parent, ignore=profile.path
    )
    iteration = 1
    while profile.profile_id in ids and iteration < 100:
        profile.profile_id = str(uuid.uuid4())
        iteration += 1



# pylint: disable=R0902,C0103
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib

import octobot_commons.databases.implementations.db_writer_reader as db_writer_reader
import octobot_commons.databases.implementations._exchange_database as _exchange_database
import octobot_commons.enums as enums


class MetaDatabase:
    def __init__(self, run_dbs_identifier, with_lock=False, cache_size=None):
        self.run_dbs_identifier = run_dbs_identifier
        self.with_lock = with_lock
        self.cache_size = cache_size
        self.database_adaptor = self.run_dbs_identifier.database_adaptor
        self.run_db: db_writer_reader.DBWriterReader = None
        self.backtesting_metadata_db: db_writer_reader.DBWriterReader = None
        self.exchange_dbs = {}

    def get_run_db(self):
        """
        :return: the run database. Opens it if not open already
        """
        if self.run_db is None:
            self.run_db = self.get_db(
                self.run_dbs_identifier.get_run_data_db_identifier()
            )
        return self.run_db

    def get_backtesting_metadata_db(self):
        """
        :return: the backtesting metadata database. Opens it if not open already
        """
        if self.backtesting_metadata_db is None:
            self.backtesting_metadata_db = self.get_db(
                self.run_dbs_identifier.get_backtesting_metadata_identifier()
            )
        return self.backtesting_metadata_db

    async def get_backtesting_metadata_from_run(self):
        """
        :return: the backtesting metadata for the associated run_dbs_identifier's backtesting_id
        """
        db = self.get_backtesting_metadata_db()
        return (
            await db.select(
                enums.CacheDatabaseTables.METADATA.value,
                (await db.search()).id == self.run_dbs_identifier.backtesting_id,
            )
        )[0]

    def _get_exchange_db(self, exchange=None):
        """
        :return: the ExchangeDatabase associated to the given exchange
        """
        exchange = exchange or self.run_dbs_identifier.context.exchange_name
        try:
            return self.exchange_dbs[exchange]
        except KeyError:
            self.exchange_dbs[exchange] = _exchange_database.ExchangeDatabase(
                self, exchange
            )
        return self.exchange_dbs[exchange]

    def get_orders_db(self, account_type, exchange=None):
        """
        :return: the orders database. Opens it if not open already
        """
        return self._get_exchange_db(exchange).get_orders_db(account_type)

    def get_trades_db(self, account_type, exchange=None):
        """
        :return: the trades database. Opens it if not open already
        """
        return self._get_exchange_db(exchange).get_trades_db(account_type)

    def get_transactions_db(self, account_type, exchange=None):
        """
        :return: the transactions database. Opens it if not open already
        """
        return self._get_exchange_db(exchange).get_transactions_db(account_type)

    def get_historical_portfolio_value_db(self, account_type, exchange):
        """
        :return: the historical portfolio database. Opens it if not open already
        """
        return self._get_exchange_db(exchange).get_historical_portfolio_value_db(
            account_type
        )

    def get_symbol_db(self, exchange, symbol):
        """
        :return: the symbol database. Opens it if not open already
        """
        return self._get_exchange_db(exchange).get_symbol_db(symbol)

    async def get_all_symbol_dbs(self, exchange):
        """
        :return: an iterable over each symbol database for the given exchange
        """
        return await self._get_exchange_db(exchange).get_all_symbol_dbs()

    def all_basic_run_db(self, account_type, exchange=None):
        """
        yields the run, orders, trades and transactions databases
        """
        yield self.get_run_db()
        exchange = exchange or self.run_dbs_identifier.context.exchange_name
        for db in self.exchange_dbs[exchange].all_basic_run_db(account_type):
            yield db

    def get_db(self, db_identifier):
        """
        :return: the database associated to the given identifier
        """
        return db_writer_reader.DBWriterReader(
            db_identifier,
            with_lock=self.with_lock,
            cache_size=self.cache_size,
            database_adaptor=self.database_adaptor,
            enable_storage=self.run_dbs_identifier.enable_storage,
        )

    async def close(self):
        """
        Closes all the open databases
        """
        # avoid asyncio.gather here as it is producing unexplained side effects (frozen thread preventing stop)
        for coro in (
            db.close()
            for db in (
                self.run_db,
                self.backtesting_metadata_db,
            )
            if db is not None
        ):
            await coro
        for exchange_db in self.exchange_dbs.values():
            await exchange_db.close()

    @classmethod
    @contextlib.asynccontextmanager
    async def database(cls, database_manager, with_lock=False, cache_size=None):
        """
        Created a local meta database and closes it upon leaving the context manager
        """
        meta_db = None
        try:
            meta_db = MetaDatabase(
                database_manager, with_lock=with_lock, cache_size=cache_size
            )
            yield meta_db
        finally:
            if meta_db is not None:
                await meta_db.close()
